
RDATATYPES = ["-"]*65535
RDATATYPES[0] = "NONE"
RDATATYPES[1] = "A"
RDATATYPES[2] = "NS"
RDATATYPES[3] = "MD"
RDATATYPES[4] = "MF"
RDATATYPES[5] = "CNAME"
RDATATYPES[6] = "SOA"
RDATATYPES[7] = "MB"
RDATATYPES[8] = "MG"
RDATATYPES[9] = "MR"
RDATATYPES[10] = "NULL"
RDATATYPES[11] = "WKS"
RDATATYPES[12] = "PTR"
RDATATYPES[13] = "HINFO"
RDATATYPES[14] = "MINFO"
RDATATYPES[15] = "MX"
RDATATYPES[16] = "TXT"
RDATATYPES[17] = "RP"
RDATATYPES[18] = "AFSDB"
RDATATYPES[19] = "X25"
RDATATYPES[20] = "ISDN"
RDATATYPES[21] = "RT"
RDATATYPES[22] = "NSAP"
RDATATYPES[23] = "NSAP_PTR"
RDATATYPES[24] = "SIG"
RDATATYPES[25] = "KEY"
RDATATYPES[26] = "PX"
RDATATYPES[27] = "GPOS"
RDATATYPES[28] = "AAAA"
RDATATYPES[29] = "LOC"
RDATATYPES[30] = "NXT"
RDATATYPES[33] = "SRV"
RDATATYPES[35] = "NAPTR"
RDATATYPES[36] = "KX"
RDATATYPES[37] = "CERT"
RDATATYPES[38] = "A6"
RDATATYPES[39] = "DNAME"
RDATATYPES[41] = "OPT"
RDATATYPES[42] = "APL"
RDATATYPES[43] = "DS"
RDATATYPES[44] = "SSHFP"
RDATATYPES[45] = "IPSECKEY"
RDATATYPES[46] = "RRSIG"
RDATATYPES[47] = "NSEC"
RDATATYPES[48] = "DNSKEY"
RDATATYPES[49] = "DHCID"
RDATATYPES[50] = "NSEC3"
RDATATYPES[51] = "NSEC3PARAM"
RDATATYPES[52] = "TSLA"
RDATATYPES[53] = "SMIMEA"
RDATATYPES[55] = "HIP"
RDATATYPES[56] = "NINFO"
RDATATYPES[59] = "CDS"
RDATATYPES[60] = "CDNSKEY"
RDATATYPES[61] = "OPENPGPKEY"
RDATATYPES[62] = "CSYNC"
RDATATYPES[64] = "SVCB"
RDATATYPES[65] = "HTTPS"
RDATATYPES[99] = "SPF"
RDATATYPES[103] = "UNSPEC"
RDATATYPES[108] = "EUI48"
RDATATYPES[109] = "EUI64"
RDATATYPES[249] = "TKEY"
RDATATYPES[250] = "TSIG"
RDATATYPES[251] = "IXFR"
RDATATYPES[252] = "AXFR"
RDATATYPES[253] = "MAILB"
RDATATYPES[254] = "MAILA"
RDATATYPES[255] = "ANY"
RDATATYPES[256] = "URI"
RDATATYPES[257] = "CAA"
RDATATYPES[258] = "AVC"
RDATATYPES[259] = "AMTRELAY"
RDATATYPES[32768] = "TA"
RDATATYPES[32769] = "DLV"