
RCODES = ["-"]*4095
RCODES[0] = "NOERROR"
RCODES[1] = "FORMERR"
RCODES[2] = "SERVFAIL"
RCODES[3] = "NXDOMAIN"
RCODES[4] = "NOIMP"
RCODES[5] = "REFUSED"
RCODES[6] = "YXDOMAIN"
RCODES[7] = "YXRRSET"
RCODES[8] = "NXRRSET"
RCODES[9] = "NOTAUTH"
RCODES[10] = "NOTZONE"
RCODES[11] = "DSOTYPENI"
RCODES[16] = "BADSIG"
RCODES[17] = "BADKEY"
RCODES[18] = "BADTIME"
RCODES[19] = "BADMODE"
RCODES[20] = "BADNAME"
RCODES[21] = "BADALG"
RCODES[22] = "BADTRUNC"
RCODES[23] = "BADCOOKIE"